/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef CDCLSTAFF_HPP_
#define CDCLSTAFF_HPP_
#define solvertrooper_hpp_version "$Revision: 1.11 $"

#include "GlobalSettings.hpp"
#include "BCPEngine.hpp"
#include "AtomicTypes.hpp"
#include "VariableInfos.hpp"
#include "Logging.hpp"
#include "Heap.hpp"
#include "BitVector.hpp"
#include "SiblingInfo.hpp"
#include "RestartStrategies.hpp"
#include "Decision.hpp"
#include "Satistics.hpp"




using std::pair; using std::make_pair;
using std::vector;

#define TROOPERTEMP template<class BCPT>
#define TROOPERCLASS SolverTrooper<BCPT>


TROOPERTEMP
struct SolverTrooper: public BCPT{
	typedef BCPT						super;
	ConflictsAndGC						conflicts;

protected:
	Restarts							restart;
	uint32_t							next_clear_earliest;
	volatile uint32_t					timeout;

	//clear clauses
	inline TrBool 	simplify();

	inline TrBool	search();

	inline bool 	enqueue(Literal lit);

	bool			import(){return  super::import();}
	//bool			import(){return super::import(((super::sstate.s_info->max_lbd_gc * 3) +5)/ 4);}
	//bool			import(){return super::import(super::sstate.s_info->max_lbd_gc +2,75);}
	//bool			import(){return super::import(3);}

	//inline void		init(); //in parallel mode
	inline SolverTrooper(Satistics* stat,uint32_t to,typename super::decision_t* dec, TrooperSiblingInfo* sinf,
			Restarts restart_init,const ConflictsAndGC& confl,const bool glue,
			const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub);
	~SolverTrooper(){};
public:
	inline void 	createClone(SolvState<SimplyFire,LeaderSiblingInfo>& st, const vector<Literal>& units); //before parallel start
	static inline SolverTrooper* create(const uint32_t& grp_sz, const uint32_t& t_id, const UserBit& leader_bit,ElimClses& eli,
			uint32_t seed, typename BCPT::decision_t::act_t max,typename BCPT::decision_t::act_t decay,ImportCriteria imprt,
			uint32_t rand_freq_permil,PolarityMode m,Restarts restart_init,
			const ConflictsAndGC& confl,uint32_t to,const bool glue,
			const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub);

	static SolverTrooper* destruct(SolverTrooper* s){
			Satistics* st=s->super::sstate.stats;	TrooperSiblingInfo* sinf=s->super::sstate.s_info;
				typename super::decision_t* dec	=s->super::decision;
				delete s;  delete dec; delete sinf; delete st; return NULL;}

	inline TrBool 	isSat();

	//inline bool		addValidSharedClause(SharedClause* cls);
	inline void		printModel(){super::printModel();}
	inline void		stopASAP();

};



TROOPERTEMP
void	TROOPERCLASS::stopASAP(){
	super::sstate.stats->nbr_confl = conflicts.count();
	timeout = 0;
	restart	.stopASAP();
}


TROOPERTEMP
TROOPERCLASS::SolverTrooper(Satistics* stat,uint32_t to, typename super::decision_t* dec, TrooperSiblingInfo* sinf,
			Restarts restart_init,const ConflictsAndGC& confl,const bool glue,
			const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub):
	super						(stat,dec,sinf,glue, beyup,hbres,otfsub),
	conflicts					(confl),
	restart						(restart_init),
	next_clear_earliest			(0),
	timeout						(to)
{}


TROOPERTEMP
void TROOPERCLASS::createClone(SolvState<SimplyFire,LeaderSiblingInfo>& st, const vector<Literal>& units){
	super::createClone(st,units);

	conflicts			.setNumberClauses(super::sstate.stats->nAllClauses());
	LOG3("Unit clauses finally inited");
}

TROOPERTEMP
TROOPERCLASS* TROOPERCLASS::create(const uint32_t& grp_sz, const uint32_t& t_id, const UserBit& leader_bit,ElimClses& eli,
			uint32_t seed,	typename BCPT::decision_t::act_t max,typename BCPT::decision_t::act_t decay,ImportCriteria imprt,
			uint32_t rand_freq_permil,PolarityMode m,
			Restarts restart_init,const ConflictsAndGC& confl,uint32_t to,const bool glue,
			const BeyondUP beyup,const HyperBinRes hbres,const OnTheFlySubsum otfsub){
	Satistics* stats 				= new Satistics();
	TrooperSiblingInfo* sinf		= new TrooperSiblingInfo(stats,grp_sz,t_id,leader_bit,imprt,eli);
	typename super::decision_t* dec	= new typename super::decision_t(stats,max,decay,seed,rand_freq_permil,m);

	SolverTrooper* ret 				=
		new SolverTrooper(stats,to,dec,sinf,restart_init,confl,glue,beyup,hbres,otfsub);
	sinf->initVarInfo(&ret->sstate.vinf);
	return ret;
}








TROOPERTEMP
TrBool 	TROOPERCLASS::isSat(){
	LOG1("random freq " << super::decision->getRandFreq());
	//LOG_BASIC("polarity mode " << m);
	LOG1("Timeout " << timeout);
	LOG1("restart init " << restart.getInitialNumber());
	LOG1("beyond Unit Prop " << super::bey_up);
	LOG1("bump_glue " << super::bump_glue);


	if(!super::ok())		return FALSE_TrBool;
	super::decision			->validateNbrVariables(super::sstate.stats->nVariables());
	super::sstate.stats->init_time = Satistics::cpuWallTime();
	TrBool  status			= UNKNOWN_TrBool;LOG2("Init finished");

	while(Satistics::cpuWallTime() <= timeout){

		LOG2("CPU User Time: "<< cpuUsrTime() << " [sec]");
		//search for a satisfying assignment
		++super::sstate.stats->nbr_starts;
		status		= search();
		if(status.isKnown()) break;

		super::backjumpKeepLevel(0);
		if(!import() || super::bcpLong()!=super::BCP_SUCCESS){
				return TrBool(super::setInvalid("import in main loop"));}
	}

	return status;
}

TROOPERTEMP
bool TROOPERCLASS::enqueue(Literal lit){
	asrt(super::sstate.vinf.litUkwn(lit));
	while(!super::enqueue(lit,super::decisionLevel())){
		LOG3("Could not enqueue lit " << lit.toDIMACS());
		if(super::decisionLevel()==0) return super::setInvalid("Enqueue of both pols failed");
		super::backjumpKeepLevel(0);
		lit = lit.flipped();
	}
	return true;
}





TROOPERTEMP
TrBool TROOPERCLASS::search( ){

	asrt(super::ok());
	restart.nextStart();

	TrBool			result			= UNKNOWN_TrBool;
	uint32_t		lrnts_alw;
	const uint32_t	import_force	= super::sstate.s_info->getImportFreqForced();
	uint32_t to_import = 0;

	while(restart.shallRun()){ // || super::decisionLevel() == 0 ){ - can cause loops

		LOG3("New Round start bcp ");
		int32_t bcp_res	= super::bcpLong();
		if(super::bey_up>NoBUP && bcp_res == super::BCP_SUCCESS){
			bcp_res 	= super::bcpBeyondUP();
			if(bcp_res == super::BCP_SUCCESS && super::openBCPImpls()) continue;
		}

		if(bcp_res == BCPT::BCP_SUCCESS){		//there is no conflict
			LOG3("Prop no conflict at level " << super::decisionLevel());

			lrnts_alw = conflicts.shallDoGC(super::sstate.stats->nLearntLongClses(),super::sstate.stats->nLits());// +super::nAssignments();
			if(lrnts_alw != max32 ){
				if(!import()) 				return TrBool(super::setInvalid("bad import before GC"));
				if(super::openBCPImpls())	continue;
				super::clsGC(lrnts_alw);
				if(!import()) 				return TrBool(super::setInvalid("bad import after GC"));
				conflicts.didGC(super::sstate.stats->nLearntLongClses());
				next_clear_earliest			= super::sstate.stats->nbr_bcp + (super::sstate.stats->nLits()/super::sstate.s_info->innerGroupSize());
				to_import					= 0;
			}

			//if(!updateGraph()) 									return TrBool(super::setInvalid("bad graph update"));
			if(super::openBCPImpls()|| super::openCompImpls())	continue;

			if (super::decisionLevel() == 0){
				to_import = 0;
				result 		= simplify();
				if(result.isKnown()) return result;
				asrt(!super::openBCPImpls());
				if(super::openCompImpls())	continue;
			}else if(++to_import > import_force){
				to_import=0;
				if(!import()){ return TrBool(super::setInvalid("bad import forced in search"));}
				if(super::openBCPImpls() || super::openCompImpls()) continue;
			}

			Literal next	= super::decision->decide();

			if (next == INVALID_LIT)	{LOG1("All variables assigned");
				return TRUE_TrBool;   }

			asrt(!super::sstate.vinf.litRepl(next));asrt(!super::sstate.vinf.varElim(next.toVar()));
			// Increase decision level and enqueue 'next'
			asrt(super::sstate.vinf.litUkwn(next));
			super::newDecisionLevel();
			if(!enqueue(next)) return TrBool(super::setInvalid("Contradicting decision"));

		}else{						//there is a conflict
			conflicts.incConflict();
			LOG3("Prop caused "<<conflicts.count() <<". conflict at level " << super::decisionLevel());
			if(!super::handleConflict(bcp_res)){
				return TrBool(super::setInvalid("Conflict at level 0")); }

			if(restart.incConflictBreak(super::confl_levs)) break;
		}
	}

	//maximal number of conflicts reached
	return UNKNOWN_TrBool;
}




TROOPERTEMP
TrBool TROOPERCLASS::simplify( ){

	LOG1("Start simplify");
	asrt(super::decisionLevel() == 0);
	super::sendUnits();

	if (!super::ok() || (super::bcpLong() != BCPT::BCP_SUCCESS)){
		return TrBool(super::setInvalid("Simplification"));}

	if(!import() || (super::bcpLong()!=super::BCP_SUCCESS)) {
		return TrBool(super::setInvalid("Bad import at simplify"));}


	if(super::sstate.stats->nbr_bcp < next_clear_earliest){return UNKNOWN_TrBool;}
	next_clear_earliest	= super::sstate.stats->nbr_bcp + (super::sstate.stats->nLits()/super::sstate.s_info->innerGroupSize());

	/*if(simp_last_idx == super::assign_stack.size() || (super::sstate.stats->nbr_bcp <= simp_next_at)){ return UNKNOWN_TrBool;}

	simp_last_idx 	= super::assign_stack.size();
	simp_next_at 	= super::sstate.stats->nbr_bcp + super::sstate.stats->nLits();*/

	asrt(!super::openBCPImpls());
	++super::sstate.stats->nbr_simplify;

	super::clearWatches(max32);
	if(!import()  || (super::bcpLong()!=super::BCP_SUCCESS)) {
				return TrBool(super::setInvalid("Bad import at simplify"));}

	LOG1("Finish simplify");
	return UNKNOWN_TrBool;
}


#endif /*CDCLSTAFF_HPP_*/

